/* CS 472/672 Assignment #2 - dirsolver.c file */

#include <stdio.h>   
#include <malloc.h>
#include "dirsolver.h"


/*                       */
/* Numeric factorization */
/*                       */
void factor(int N, int *ia, int *ja, double *a, int *order, int *invord, struct sparse_row *LU) 
/*
   input:

      N       number of unknowns
      ia, ja  stucture of a
      a       real values of a
      order   order[new_order] = old_order
      invord  invord[old_order] = new_order
      LU      contains info about symbolic factors of the LU 

   output:

      LU      now contains real LU as well as symbolic 
*/
{
 
   printf("You have to write the function factor.\n");

}



/*                */
/* Solve LU x = b */
/*                */
void solve(int N, double *x, double *b, int *order, struct sparse_row *LU)
/*
   input:

      N      number of unknowns
      b      rhs
      order  order[new_order] = old_order
      LU     sparse_row structure containing LU

   output:

      x      solution in the original order
*/
{

   printf("You have to write the function solve.\n");

}



/* ******************************************************************* */
/*               Analyze phase                                         */
/*               (You should not change anything below)                */
/* ******************************************************************* */


/*                      */
/* Merge rows           */
/* (symmetric approach) */
/*                      */
void  merge_smart(int i, int *list, int N, int *link, struct sparse_row *LU) 
/*
   input:

      i     current row being factored (Dolittle form)
      list  implied linked list of nonzeros in current
            row.  On entry, only original nonzero columsn
            in this row.  1st nonzero column = first,
            list[first], list[ list[first]], ...., until
            list[  [....] ] = N+1 -> signals end of list
      N     number of unknowns
      LU    ptr to sparse row structure containing info about
            previous rows

      link  implied linked list of rows whose first non-zero is in col i, i.e.
            list of prev rows whose first nonzero col is in col i given by
            link[i], link[link[i]], ..., until link[  [....] ] = N+1 
            -> signals end of list

   output:

     list   now contains all nonzeros in this factored row i
            (implied linked list)
*/
{
   int row, next, oldlst, ii;
   int nxtlst;


   next = link[i];

   while (next != N+1) {

      oldlst = i;
      nxtlst = list[i];
      row = next;
 
      /* scan row "row" of U */
      for (ii=LU[row].diag+1; ii<LU[row].nz; ii++) {
 
         /* skip thru linked list until we find possible insertion pt */
         while (nxtlst<LU[row].jaf[ii]) {
            oldlst = nxtlst;
            nxtlst = list[oldlst]; 
         }
                
         if (nxtlst>LU[row].jaf[ii]) {
            list[oldlst] = LU[row].jaf[ii];
            list[LU[row].jaf[ii]] = nxtlst;
            oldlst = LU[row].jaf[ii];
         }
         else {
            oldlst = nxtlst;
            nxtlst = list[oldlst];
         }
      }
      next = link[next];

   }

}


/*                        */
/* Symbolic factorization */
/* (symmetric approach)   */
/*                        */
void sfac_smart(int *ia, int *ja, int N, int *order, int *invord, int* nnz, struct sparse_row *LU)
/* 
   input:

     ia, ja  structure of a
     N       number of unknowns
     order   order[new_order] = old_order
     invord  invord[old_order] = new_order
 
  output:

     nnz     number of nonzeros in LU
     LU      sparse_row stucture contains symbolic ILU info
*/
{
   int *int_temp, *list, *link, *count;
   int i, ii, k, iold, num, next, icol, nrow;
 

   /* Allocate temp wkspace */
   list = (int *) malloc(N*sizeof(int));
   int_temp = (int *) malloc(N*sizeof(int));
   link = (int *) malloc(N*sizeof(int));
   count = (int *) malloc(N*sizeof(int));


   /* Symbolic factorization */
   for (i=0; i<N; i++) {
      list[i] = N+1;
      link[i] = N+1;
      count[i] = 0;   /* used to keep trak of number of
                         nonzeros in a row, based on symmetry
                         since we get only struct of U from
                         smart algorithm
                      */
   }
 
   *nnz = 0;
   for (i=0; i<N; i++) {
      iold = order[i];
 
      ii = 0;
      for (k=ia[iold]; k<ia[iold+1]; k++){
         int_temp[ii] = invord[ja[k]] ;
         ii++;
      }

      /* sort entries */
      num = ia[iold+1] - ia[iold];
      if (num>1) sortq(int_temp, 0, num-1); 

      /* define implied linked list */
      for (ii=0; ii<num-1; ii++) {
         if (int_temp[ii] >= i) {
            list[ int_temp[ii] ] = int_temp[ii+1];
         }
      }
      list[int_temp[num-1]] = N+1;

      /* merge rows defined in implied linked list */
      merge_smart(i, list, N, link, LU);

      /* count up nonzeros in this row */
      nrow=0;
      next = i;
      while (next != N+1){
         nrow++;
         next = list[next];
      }
      nrow = nrow + count[i];
      LU[i].nz = nrow;
      *nnz += nrow;

      /* allocate space for this row */
      LU[i].jaf = (int *) malloc(nrow*sizeof(int));
      LU[i].af = (double *) malloc(nrow*sizeof(double));

      /* define jaf and diag. count[i]=nnz of row i of L */         
      next = i;
      k = count[i];
      while (next != N+1) {
         LU[i].jaf[k] = next;
         if (next == i) {
            LU[i].diag = k;
         }
         count[next]++;
         next = list[next];
         k++;
      }

      /* linked list of rows whose first non-zero is in col icol */
      if (LU[i].diag+1 < LU[i].nz) {
         icol = LU[i].jaf[LU[i].diag+1];
         ii = link[icol];
         link[icol] = i;
         link[i] = ii;
      }

   }


   /*  Fill in lower triangular part */
   for (i=0; i<N; i++) count[i] = 0;

   for (i=0; i<N; i++) {
      for (k=LU[i].diag+1; k<LU[i].nz; k++) {
         icol = LU[i].jaf[k];
         LU[icol].jaf[count[icol]] = i;
         count[icol]++;
      }
   }


   /* Free integer temp */
   free(int_temp); free(list); free(link); free(count);

}


/*                                                     */
/* Auxilliary functions used by symbolic factorization */
/*                                                     */

void shell(int *v, int n)
/*    
   shellsort: sort v[0], ..., v[*n_ptr-1] into increasing order
*/
{  
   int gap, i, j, temp;

   for (gap=n/2; gap>0; gap=gap/2) {
      for (i=gap; i<n; i++) {
         for (j=i-gap; j>=0 && v[j]>v[j+gap]; j=j-gap) {
            temp = v[j];
            v[j] = v[j+gap];
            v[j+gap] = temp;
         }
      }
   }
}


void swap(int *v, int i, int j)
/* 
   function used by sortq
*/
{
   int temp;

   temp = v[i];
   v[i] = v[j];
   v[j] = temp;
}


void sortq(int *v, int left, int right)
/* 
   simple quick sort 
*/
{
   int i, last, num_sort;

   num_sort = right - left + 1;

   if (num_sort<40) {
      shell( &v[left], num_sort);
      return;
   }

   if (left>=right) {
      return;
   }

   swap(v, left, (left + right)/2);

   last = left;

   for (i=left+1; i<=right; i++) {
      if (v[i]<v[left]) {
         swap(v, ++last, i);
      }
   }

   swap(v, left, last);

   sortq(v, left, last-1);
   sortq(v, last+1, right);
}


void clean(int N, struct sparse_row *LU)
/*    
   Free space for LU in after finished

   input:

     N   number of unknowns
     LU  structure containing LU (global static)

   output:

     LU  structure deallocated
*/
{
    int    i;

    for (i=0; i<N; i++) {
       free(LU[i].jaf);
       free(LU[i].af);
    }
    free (LU);
}

